import logging
import queue
import functools
from collections.abc import Callable
from typing import NoReturn

import paho.mqtt.client as mqtt
from paho.mqtt.client import topic_matches_sub

class Configuration:
    user: str
    password: str
    host: str
    port: int

class Client(mqtt.Client):
    def __init__(self, logger: logging.Logger=None):
        super().__init__()

        self.logger = logger
        if self.logger is None:
            self.logger = logging.getLogger('mqtt')

        self._on_connect_callbacks = []
        self._subscribe_callbacks = {}

        self.on_connect = self._on_connect2

    def setup(self, host: str, port: int, username: str, password: str):
        self._username = username
        self._password = password
        self._host = host
        self._port = port

    def connect(self):
        self.reconnect()

    def is_alive(self):
        """ Check for loop thread health """
        return self._thread.is_alive()

    def join(self):
        return self._thread.join()

    def subscribe_callback(self, topic: str, callback):
        self.logger.info("Add callback %s for topic %s" % (callback, topic))

        if topic in self._subscribe_callbacks:
            self._subscribe_callbacks[topic].append(callback)
        else:
            self._subscribe_callbacks[topic] = [callback]

        #register our callback for storing multiple callback for one topic
        self.message_callback_add(topic, self.on_subscribe_callback)

    def on_subscribe_callback(self, client, userdata, msg):
        self.logger.debug("Callback search for %s" % msg.topic)
        for topic, callbacks in self._subscribe_callbacks.items():
            if topic_matches_sub(topic, msg.topic):
                for func in callbacks:
                    self.logger.debug("Call %s for topic %s" % (func, topic))
                    func(client, userdata, msg)


    def _on_connect2(self, *args):
        for topic in self._subscribe_callbacks.keys():
            self.subscribe(topic)
        for func in self._on_connect_callbacks:
            func(self, *args)

    def add_on_connect_callback(self, func):
        with self._callback_mutex:
            self._on_connect_callbacks.append(func)

    @mqtt.Client.on_connect.setter
    def on_connect(self, func):
        with self._callback_mutex:
            self._on_connect = func


class MQTTApp:
    """ Manage MQTT connection and on topic callback with help of the @topic decorator """

    def __init__(self, logger : logging.Logger=None):
        """logger: specify a logger to use, default: mqttapp"""
        self.context = {}

        self._mqtt_topic_route = {}
        self._mqttc = None

        self._mqtt_host = None
        self._mqtt_port = None

        self._exception_queue = queue.Queue() #get exceptions from callback

        self.logger = logger

        if self.logger is None:
            self.logger = logging.getLogger('mqttapp')

        self.context['logger'] = self.logger

    def topic(self,mqtt_topic : str) -> Callable:
        """ A decorator for subscribe to a topic with decorated func as callback
        and catch all exception inside the func to raise it when check is_healthy
        method """
        def wrap(f):
            @functools.wraps(f)
            def wraped_f(*arg, **kwarg):
                try:
                   return f(*arg, **kwarg)
                except Exception as e:
                   self._exception_queue.put(e)
            self._mqtt_topic_route[mqtt_topic] = wraped_f
            return wraped_f
        return wrap

    def setup(self, host: str, port: int, user: str, pwd: str) -> NoReturn:
        """ Setup mqtt connection, call it before loop_start() """
        self._mqttc = mqtt.Client(clean_session=True)
        self._mqttc.username_pw_set(user, password=pwd)
        self._mqttc.on_connect = self._on_connect
        self._mqtt_host = host
        self._mqtt_port = port
        self._mqttc.user_data_set(self.context)

    def _on_connect(self, client, userdata, flags, rc):
        for topic, func in self._mqtt_topic_route.items():
            self.logger.info("Add callback %s on %s" % (func.__name__,topic))
            self._mqttc.message_callback_add(topic, func)
            self._mqttc.subscribe(topic)

    def loop_start(self) -> NoReturn:
        self._mqttc.connect(self._mqtt_host, self._mqtt_port, 60)
        self._mqttc.loop_start()

    def loop_stop(self) -> NoReturn:
        self._mqttc.loop_stop()

    def is_healthy(self) -> bool:
        """ Check for exception in callback thread,
        and raise it in main thread """
        try:
            exc = self._exception_queue.get(block=False)
        except queue.Empty:
            return True
        else:
            raise exc

    @property
    def client(self):
        return self._mqttc
